import threading
from queue import Queue
from file_writer_control.KafkaTopicUrl import KafkaTopicUrl
from kafka import KafkaConsumer
from kafka.errors import NoBrokersAvailable
from typing import List, Union, Dict
import atexit
from datetime import datetime

from file_writer_control.InThreadStatusTracker import InThreadStatusTracker
from file_writer_control.WorkerStatus import WorkerStatus, WorkerState

from file_writer_control.JobStatus import JobStatus, JobState
from file_writer_control.CommandStatus import CommandStatus, CommandState
from file_writer_control.InThreadStatusTracker import (
    STATUS_MESSAGE_TIMEOUT,
    COMMAND_STATUS_TIMEOUT,
    JOB_STATUS_TIMEOUT,
)


def thread_function(host_port: str, topic: str, in_queue: Queue, out_queue: Queue):
    """
    Background thread for consuming Kafka messages.
    :param host_port: The host + port of the Kafka broker that we are using.
    :param topic: The Kafka topic that we are listening to.
    :param in_queue: A queue for sending "exit" messages to the thread.
    .. note:: The queue will exit upon the reception of the string "exit" on this queue.
    :param out_queue: The queue to which status updates are published.
    """
    status_tracker = InThreadStatusTracker(out_queue)
    while True:
        try:
            consumer = KafkaConsumer(
                topic,
                bootstrap_servers=host_port,
                fetch_max_bytes=52428800 * 6,
                consumer_timeout_ms=100,
            )  # Roughly 300MB
            break
        except NoBrokersAvailable:
            pass  # Do not fail if the broker is not immediately available.
        if not in_queue.empty():
            new_msg = in_queue.get()
            if new_msg == "exit":
                return
    while True:
        for message in consumer:
            status_tracker.process_message(message.value)
        status_tracker.check_for_lost_connections()
        if not in_queue.empty():
            new_msg = in_queue.get()
            if new_msg == "exit":
                break
    consumer.close(True)


class CommandChannel(object):
    """
    A class that implements the functionality for receiving and interpreting messages that are published to the
    Kafka command topic of a pool of file-writers.
    .. note:: This class implements a thread that will continuously attempt to connect to a Kafka broker.
    """

    def __init__(self, command_topic_url: str):
        """
        Constructor.
        :param command_topic_url: The url of the Kafka topic to where the file-writer status/command messages are published.
        """
        kafka_address = KafkaTopicUrl(command_topic_url)
        self.status_queue = Queue()
        self.to_thread_queue = Queue()
        thread_kwargs = {
            "host_port": kafka_address.host_port,
            "topic": kafka_address.topic,
            "in_queue": self.to_thread_queue,
            "out_queue": self.status_queue,
        }
        self.map_of_workers: Dict[str, WorkerStatus] = {}
        self.map_of_jobs: Dict[str, JobStatus] = {}
        self.map_of_commands: Dict[str, CommandStatus] = {}
        self.run_thread = True
        self.thread = threading.Thread(
            target=thread_function, daemon=True, kwargs=thread_kwargs
        )
        self.thread.start()

        def do_exit():
            self.stop_thread()

        atexit.register(do_exit)

    def add_job_id(self, job_id: str):
        """
        Add a job identifier to the list of known jobs before it has been encountered on the command topic.
        :param job_id: The identifier of the new job.
        """
        if job_id not in self.map_of_jobs:
            self.map_of_jobs[job_id] = JobStatus(job_id)

    def add_command_id(self, job_id: str, command_id: str):
        """
        Add a command identifier to the list of known commands before it has been encountered on the command topic.
        :param job_id: The job identifier of the new command.
        :param command_id: The identifier of the new command.
        """
        if command_id not in self.map_of_commands:
            self.map_of_commands[command_id] = CommandStatus(job_id, command_id)

    def stop_thread(self):
        """
        Stop the thread that is continuously getting command topic messages in the background. Should only be called if
        we are about to get rid of the current instance of CommandChannel.
        """
        self.to_thread_queue.put("exit")
        try:
            self.thread.join()
        except RuntimeError:
            pass  # Do not throw an exception if the thread has not yet been started.

    def __del__(self):
        self.stop_thread()

    def update_workers(self):
        """
        Update the list of known workers, jobs and commands. This is a non-blocking call but it might take some time
        to execute if the queue of updates is long. This member function is called by many of the other member functions
        in this class.
        """

        def handle_worker_status(status_update):
            if status_update.service_id not in self.map_of_workers:
                self.map_of_workers[status_update.service_id] = status_update
            self.map_of_workers[status_update.service_id].update_status(status_update)

        def handle_job_status(status_update):
            if status_update.job_id not in self.map_of_jobs:
                self.map_of_jobs[status_update.job_id] = status_update
            self.map_of_jobs[status_update.job_id].update_status(status_update)

        def handle_command_status(status_update):
            if status_update.command_id not in self.map_of_commands:
                self.map_of_commands[status_update.command_id] = status_update
            self.map_of_commands[status_update.command_id].update_status(status_update)

        status_updater_map = {
            WorkerStatus: handle_worker_status,
            CommandStatus: handle_command_status,
            JobStatus: handle_job_status,
        }
        while not self.status_queue.empty():
            status_update = self.status_queue.get()
            status_updater_map[type(status_update)](status_update)

        now = datetime.now()
        for worker in self.map_of_workers.values():
            if worker.last_update + STATUS_MESSAGE_TIMEOUT < now:
                worker.state = WorkerState.UNAVAILABLE
        for job in self.map_of_jobs.values():
            if job.last_update + JOB_STATUS_TIMEOUT < now and not (
                job.state == JobState.DONE or job.state == JobState.ERROR
            ):
                job.state = JobState.TIMEOUT
        for command in self.map_of_commands.values():
            if command.last_update + COMMAND_STATUS_TIMEOUT < now and not (
                command.state == CommandState.SUCCESS
                or command.state == CommandState.ERROR
            ):
                command.state = CommandState.TIMEOUT_RESPONSE

    def list_workers(self) -> List[WorkerStatus]:
        """
        :return: A list of the (known) workers with state and status information.
        """
        self.update_workers()
        return list(self.map_of_workers.values())

    def list_jobs(self) -> List[JobStatus]:
        """
        :return: A list of the (known) jobs with state and status information.
        """
        self.update_workers()
        return list(self.map_of_jobs.values())

    def list_commands(self) -> List[CommandStatus]:
        """
        :return: A list of the (known) commands and their outcomes.
        """
        self.update_workers()
        return list(self.map_of_commands.values())

    def get_job(self, job_id: str) -> Union[JobStatus, None]:
        """
        Get the status of a single job.
        :param job_id: The job identifier of the job we are interested in.
        :return: The job status or None if the job is not known.
        """
        self.update_workers()
        if job_id in self.map_of_jobs:
            return self.map_of_jobs[job_id]
        return None

    def get_worker(self, service_id: str) -> Union[WorkerStatus, None]:
        """
        Get the status of a single worker.
        :param service_id: The service identifier of the worker we are interested in.
        :return: The worker status or None if the service id is not known.
        """
        self.update_workers()
        if service_id in self.map_of_workers:
            return self.map_of_workers[service_id]
        return None

    def get_command(self, command_id: str) -> Union[CommandStatus, None]:
        """
        Get the status of a single command.
        :param command_id: The command identifier of the command we are interested in.
        :return: The command status/outcome or None if the command is not known.
        """
        self.update_workers()
        if command_id in self.map_of_commands:
            return self.map_of_commands[command_id]
        return None



from queue import Queue
from streaming_data_types.status_x5f2 import FILE_IDENTIFIER as STAT_IDENTIFIER
from streaming_data_types.run_start_pl72 import FILE_IDENTIFIER as START_IDENTIFIER
from streaming_data_types.run_stop_6s4t import FILE_IDENTIFIER as STOP_TIME_IDENTIFIER
from streaming_data_types.finished_writing_wrdn import (
    FILE_IDENTIFIER as STOPPED_IDENTIFIER,
)
from datetime import datetime, timedelta
from streaming_data_types.utils import _get_schema
from streaming_data_types import deserialise_x5f2 as deserialise_status
from streaming_data_types import deserialise_answ as deserialise_answer
from streaming_data_types import deserialise_6s4t as deserialise_stop_time
from streaming_data_types import deserialise_pl72 as deserialise_start
from streaming_data_types import deserialise_wrdn as deserialise_stopped
from streaming_data_types.run_stop_6s4t import RunStopInfo
from streaming_data_types.run_start_pl72 import RunStartInfo
from streaming_data_types.status_x5f2 import StatusMessage
from streaming_data_types.finished_writing_wrdn import WritingFinished
from streaming_data_types.action_response_answ import FILE_IDENTIFIER as ANSW_IDENTIFIER
from file_writer_control.JobStatus import JobStatus, JobState
from file_writer_control.CommandStatus import CommandStatus, CommandState
from file_writer_control.WorkerStatus import WorkerState, WorkerStatus
from file_writer_control.StateExtractor import (
    extract_worker_state_from_status,
    extract_state_from_command_answer,
    extract_job_state_from_answer,
)
import json
from streaming_data_types.action_response_answ import Response
from typing import Dict

STATUS_MESSAGE_TIMEOUT = timedelta(seconds=5)
JOB_STATUS_TIMEOUT = timedelta(seconds=5)
COMMAND_STATUS_TIMEOUT = timedelta(seconds=20)


class InThreadStatusTracker:
    """
    Implements de-coding of flatbuffer messages and sends updates of worker, job and command state/status back to the
    "main"-thread if there has been changes.
    """

    def __init__(self, status_queue: Queue):
        """
        Constructor.
        :param status_queue: The output queue to which state/status updates are pushed.
        """
        self.queue = status_queue
        self.known_workers: Dict[str, WorkerStatus] = {}
        self.known_jobs: Dict[str, JobStatus] = {}
        self.known_commands: Dict[str, CommandStatus] = {}

    def process_message(self, message: bytes):
        """
        Process a binary message.
        :param message: The binary message to be processed.
        """
        current_schema = _get_schema(message).encode("utf-8")
        update_time = datetime.now()
        msg_process_map = {
            ANSW_IDENTIFIER: lambda msg: self.process_answer(deserialise_answer(msg)),
            STAT_IDENTIFIER: lambda msg: self.process_status(deserialise_status(msg)),
            STOP_TIME_IDENTIFIER: lambda msg: self.process_set_stop_time(
                deserialise_stop_time(msg)
            ),
            START_IDENTIFIER: lambda msg: self.process_start(deserialise_start(msg)),
            STOPPED_IDENTIFIER: lambda msg: self.process_stopped(
                deserialise_stopped(msg)
            ),
        }
        if current_schema in msg_process_map:
            msg_process_map[current_schema](message)

        self.send_status_if_updated(update_time)

    def send_status_if_updated(self, limit_time: datetime):
        """
        Sends status updates of workers, jobs and commands (to the status queue) if there has been any updates on or
        after the limit_time.
        :param limit_time: The cut-off time for deciding which updates should be sent to the status queue.
        """
        for worker in self.known_workers.values():
            if worker.last_update >= limit_time:
                self.queue.put(worker)
        for job in self.known_jobs.values():
            if job.last_update >= limit_time:
                self.queue.put(job)
        for command in self.known_commands.values():
            if command.last_update >= limit_time:
                self.queue.put(command)

    def check_for_worker_presence(self, service_id: str):
        """
        Check if a service_id is known and add it to a list of known ones if it is not.
        :param service_id: The service identifier to look for.
        """
        if service_id not in self.known_workers:
            self.known_workers[service_id] = WorkerStatus(service_id)

    def check_for_job_presence(self, job_id: str):
        """
        Check if a job identifier is known and add it to a list of known ones if it is not.
        :param job_id: The job identifier to look for.
        """
        if job_id not in self.known_jobs:
            new_job = JobStatus(job_id)
            self.known_jobs[job_id] = new_job

    def check_for_command_presence(self, job_id: str, command_id: str):
        """
        Check if a command identifier is known and add it to a list of known ones if it is not.
        :param job_id: The job identifier of the command that we are looking for. (Only used if we need to add the
        command.)
        :param command_id: The command identifier to look for.
        """
        if command_id not in self.known_commands:
            new_command = CommandStatus(job_id, command_id)
            self.known_commands[command_id] = new_command

    def check_for_lost_connections(self):
        """
        Check workers, commands and jobs for the last update time and change the state of these if a timeout has been
        reached.
        """
        now = datetime.now()
        for worker in self.known_workers.values():
            if (
                worker.state != WorkerState.UNAVAILABLE
                and now - worker.last_update > STATUS_MESSAGE_TIMEOUT
            ):
                worker.state = WorkerState.UNAVAILABLE
        for command in self.known_commands.values():
            if (
                command.state != CommandState.SUCCESS
                and command.state != CommandState.ERROR
                and now - command.last_update > COMMAND_STATUS_TIMEOUT
            ):
                command.state = CommandState.TIMEOUT_RESPONSE
        for job in self.known_jobs.values():
            if (
                job.state != JobState.DONE
                and job.state != JobState.ERROR
                and now - job.last_update > JOB_STATUS_TIMEOUT
            ):
                job.state = JobState.TIMEOUT

    def process_answer(self, answer: Response):
        """
        Update workers, jobs and commands based on information in a response message.
        :param answer: The response/answer message to use for status updates.
        """
        self.check_for_worker_presence(answer.service_id)
        self.check_for_job_presence(answer.job_id)
        self.check_for_command_presence(answer.job_id, answer.command_id)
        new_job_state = extract_job_state_from_answer(answer)
        if new_job_state is not None:
            self.known_jobs[answer.job_id].state = new_job_state
        self.known_commands[
            answer.command_id
        ].state = extract_state_from_command_answer(answer)
        self.known_commands[answer.command_id].message = answer.message
        self.known_jobs[answer.job_id].message = answer.message

    def process_status(self, status_update: StatusMessage):
        """
        Update workers and jobs based on information in a status message.
        :param status_update: The status message to use for updates.
        """
        self.check_for_worker_presence(status_update.service_id)
        current_state = extract_worker_state_from_status(status_update)
        self.known_workers[status_update.service_id].state = current_state
        if current_state == WorkerState.WRITING:
            job_id = json.loads(status_update.status_json)["job_id"]
            self.check_for_job_presence(job_id)
            self.known_jobs[job_id].state = JobState.WRITING
            # For some jobs, we will only know the service-id when a worker starts working on a job.
            # Thus we need the following statement to update the (known) service-id of a job.
            try:
                self.known_jobs[job_id].service_id = status_update.service_id
            except RuntimeError:
                pass  # Expected error (i.e. the job is not known), do nothing

    def process_set_stop_time(self, stop_time: RunStopInfo):
        """
        Update commands and jobs based on information in a "set stop time" message.
        :param stop_time: The "stop" message to use for updates.
        """
        self.check_for_command_presence(stop_time.job_id, stop_time.command_id)
        self.known_commands[stop_time.command_id].state = CommandState.WAITING_RESPONSE

    def process_start(self, start: RunStartInfo):
        """
        Update commands and jobs based on information in a "start" message.
        :param start: The "start" message to use for updates.
        """
        self.check_for_job_presence(start.job_id)
        self.check_for_command_presence(start.job_id, start.job_id)
        self.known_commands[start.job_id].state = CommandState.WAITING_RESPONSE

    def process_stopped(self, stopped: WritingFinished):
        """
        Update workers and jobs based on information in a "has stopped" message.
        :param stopped: The "stopped" message to use for updates.
        """
        self.check_for_job_presence(stopped.job_id)
        self.check_for_worker_presence(stopped.service_id)
        current_job = self.known_jobs[stopped.job_id]
        if stopped.error_encountered:
            current_job.state = JobState.ERROR
        else:
            current_job.state = JobState.DONE
        current_job.message = stopped.message
        self.known_workers[stopped.service_id].state = WorkerState.IDLE



from file_writer_control.CommandChannel import CommandChannel
from file_writer_control.CommandStatus import CommandState


class CommandHandler:
    """
    A stand in for (more easily) checking the state of a command sent to a file-writer.
    """

    def __init__(self, command_channel: CommandChannel, command_id: str):
        """
        Constructor.
        :param command_channel: The instance of a CommandChannel that this class uses for getting the command status from.
        :param command_id: The (unique) command identifier.
        """
        self.command_id = command_id
        self.command_channel = command_channel

    def get_state(self) -> CommandState:
        """
        Get the command state.
        :return: The state of the command if possible. CommandState.UNKNOWN if the the state can not be determined.
        """
        command = self.command_channel.get_command(self.command_id)
        if command is None:
            return CommandState.UNKNOWN
        return command.state

    def is_done(self) -> bool:
        """
        :return: True if the command completed successfully. False otherwise.
        """
        return (
            self.command_channel.get_command(self.command_id).state
            == CommandState.SUCCESS
        )

    def get_message(self) -> str:
        """
        :return: If there was an error executing the command, this member function will return the error string as
        sent by the file-writer. Will return an empty string otherwise.
        """
        command = self.command_channel.get_command(self.command_id)
        if command is None:
            return ""
        return command.message



import re


class KafkaTopicUrl:
    """
    Class for extracting address, port and topic name from a Kafka topic url.
    """

    test_regexp = re.compile(
        r"^\s*(?:kafka://)?(?:(?P<host>[^/?#:]+)(?::(?P<port>\d+){1,5})?)/(?P<topic>[a-zA-Z0-9._-]+)\s*$"
    )

    def __init__(self, url: str):
        result = re.match(KafkaTopicUrl.test_regexp, url)
        if result is None:
            raise RuntimeError("Unable to match kafka url.")
        self.port = 9092  # Default Kafka broker port
        if result.group("port") is not None:
            self.port = int(result.group("port"))
        self.host = result.group("host")
        self.host_port = f"{self.host}:{self.port}"
        self.topic = result.group("topic")
